import argparse

from arch_blueprint.blueprint import ArchBlueprint


def main() -> None:
    """Main entry point for the arch_blueprint CLI."""
    parser = argparse.ArgumentParser(
        description="Generate component diagrams in plantuml for python applications",
    )
    parser.add_argument(
        "root",
        type=str,
        help="Name of root python module in project (example: 'myapp'",
    )
    parser.add_argument(
        "--modules",
        "-m",
        type=str,
        nargs="*",
        action="extend",
        help=(
            "Selected modules for rendering "
            "(examples: 'myapp.somemodule', "
            "'myapp.somemodule.*', 'myapp.somemodule.**')"
        ),
    )
    args = parser.parse_args()

    ArchBlueprint(root=args.root, target_names=args.modules).run()


main()



import textwrap
from collections.abc import Sequence
from typing import Optional

from arch_blueprint.modules import BlueprintModule


class PlantUmlRenderer:
    """Renders PlantUML diagrams from blueprint modules."""

    def __init__(self, colors: Optional[Sequence[str]] = None) -> None:
        self.colors = colors or [
            "#E74C3C",
            "#3498DB",
            "#2ECC71",
            "#1ABC9C",
            "#F39C12",
            "#9B59B6",
            "#27AE60",
            "#34495E",
            "#E67E22",
            "#8E44AD",
        ]

    def render(self, target_modules: list[BlueprintModule]) -> None:
        header = textwrap.dedent("""\
            @startuml
            !theme amiga

            top to bottom direction
            hide empty members

            """)

        body = self._render_classes(target_modules)
        body += self._render_links(target_modules)

        footer = textwrap.dedent("""\
            @enduml
        """)

        text = header + body + footer
        print(text)  # noqa: T201

    def _render_classes(self, blueprint_modules: list[BlueprintModule]) -> str:
        class_lines = []
        for blueprint_module in blueprint_modules:
            color = self.generate_color_code(blueprint_module.name)
            text = f"class {blueprint_module.name} <<(M, {color})>>\n"
            class_lines.append(text)

        return "\n".join(class_lines)

    def _render_links(self, blueprint_modules: list[BlueprintModule]) -> str:
        links = set()
        for blueprint_module in blueprint_modules:
            _links = blueprint_module.find_dependencies_namespace_to_namespaces()
            for link in _links:
                links.add(link)

        text = ""
        arrow = "--->"
        for from_, to_ in links:
            text += f"{from_} {arrow} {to_}\n"

        return text

    def generate_color_code(self, module: str) -> str:
        depth = len(module.rsplit("."))
        return self.colors[depth]



from collections.abc import Sequence
from typing import Optional

import grimp

from arch_blueprint.modules import BlueprintModule
from arch_blueprint.puml import PlantUmlRenderer


class ArchBlueprint:
    """Generates architecture blueprints for Python applications."""

    def __init__(
        self,
        root: str,
        target_names: Sequence[str],
        renderer: Optional[PlantUmlRenderer] = None,
    ) -> None:
        self.root = root
        self.target_names = target_names
        self.graph = grimp.build_graph(self.root)
        self.renderer = renderer or PlantUmlRenderer()

    def run(self) -> None:
        blueprint_modules = self.collect_modules()
        self.renderer.render(blueprint_modules)

    def collect_modules(self) -> list[BlueprintModule]:
        module_names = self.prepare_modules_list()
        result = []
        for name in module_names:
            module = self.build_module(name, module_names)
            result.append(module)

        return result

    def prepare_modules_list(self) -> set[str]:
        module_names: set[str] = set()
        for name in self.target_names:
            modules = self.graph.find_matching_modules(name)
            module_names.update(modules)
        return self._exclude_sub_modules(module_names)

    def build_module(self, name: str, module_names: set[str]) -> BlueprintModule:
        dependencies = self._find_all_modules_imported_by(name)
        return BlueprintModule(
            name=name,
            dependencies=dependencies,
            selected_modules=module_names,
        )

    def _find_all_modules_imported_by(self, module: str) -> set[str]:
        result = set()
        descends = self.graph.find_descendants(module)
        if not descends:
            return self.graph.find_modules_directly_imported_by(module)

        for descend in descends:
            imported_mods = self.graph.find_modules_directly_imported_by(descend)
            result.update(imported_mods)

        return result

    @staticmethod
    def _exclude_sub_modules(modules: set[str]) -> set[str]:
        """Filter out names that are namespaces of other modules in the set."""
        sorted_names = sorted(modules, key=len, reverse=True)
        result: set[str] = set()

        for name in sorted_names:
            is_namespace = False
            for longer_name in list(result):
                if name in longer_name:
                    is_namespace = True
                    break

            if not is_namespace:
                result.add(name)

        return result



from dataclasses import dataclass
from typing import Tuple


@dataclass
class BlueprintModule:
    """Represents a module in the architecture blueprint."""

    name: str
    dependencies: set[str]
    selected_modules: set[str]

    def get_namespace(self) -> str:
        return self.get_namespace_of_module(self.name)

    def find_dependencies_namespace_to_namespaces(self) -> set[Tuple[str, str]]:
        res = set()

        for dep in self.dependencies:
            for selected_module in self.selected_modules:
                if selected_module in dep:
                    from_, to_ = self.extract_namespaces_with_same_depth(dep)
                    if from_ != to_:
                        res.add((from_, to_))

        return res

    def is_same_namespace(self, other: str) -> bool:
        namespace = self.get_namespace_of_module(other)
        return self.get_namespace() == namespace

    def get_namespace_of_module(self, module: str) -> str:
        namespace, _ = module.rsplit(".", maxsplit=1)
        return namespace

    def extract_namespaces_with_same_depth(self, module: str) -> tuple[str, str]:
        from_ = self.name.split(".")
        to_ = module.split(".")

        path_from = []
        path_to = []
        for first, second in zip(from_, to_):
            path_from.append(first)
            path_to.append(second)
            if first != second:
                break

        return ".".join(path_from), ".".join(path_to)
